from .heartpy import *


__name__ = "HeartPy"
__author__ = "Paul van Gent"
__version__ = "Version 1.2.6"
__license__ = "GNU General Public License V3.0"


#module level docstring
__doc__ = '''
Noise-resistant heart rate analysis module for Python
=====================================================

HeartPy started out as a Python script to detect beats in
PPG heart rate data streams. It has since grown into a module capable 
of pre-processing and analysing PPG and ECG data. 


References:
-----------
van Gent, P., Farah, H., van Nes, N., & van Arem, B. (2018). 
Heart Rate Analysis for Human Factors: Development and Validation of an 
Open Source Toolkit for Noisy Naturalistic Heart Rate Data. 
In Proceedings of the 6th HUMANIST Conference (pp. 173 to 178)

van Gent, P., Farah, H., van Nes, N., & van Arem, B. (2019). 
Analysing Noisy Driver Physiology Real-Time Using Off-the-Shelf Sensors: 
Heart Rate Analysis Software from the Taking the Fast Lane Project. 
Journal of Open Research Software, 7(1), 32. 
DOI: http://doi.org/10.5334/jors.241


Documentation
-------------
https://python-heart-rate-analysis-toolkit.readthedocs.io/en/latest/
'''